package com.sun.java.util.jar.pack;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.Arrays;

final class Histogram
{
    protected final int[][] matrix;
    protected final int totalWeight;
    protected final int[] values;
    protected final int[] counts;
    private static final long LOW32 = 4294967295L;
    private static double log2;
    private final Histogram.BitMetric bitMetric;
    
    public Histogram(int[] array) {
        this.bitMetric = new Histogram.BitMetric() {
            @Override
            public double getBitLength(int n) {
                return Histogram.this.getBitLength(n);
            }
        };
        long[] computeHistogram2Col = computeHistogram2Col(maybeSort(array));
        int[][] table = makeTable(computeHistogram2Col);
        this.values = table[0];
        this.counts = table[1];
        this.matrix = makeMatrix(computeHistogram2Col);
        this.totalWeight = array.length;
        assert !(!this.assertWellFormed(array));
        return;
    }
    
    public Histogram(int[] array, int n, int n2) {
        this(sortedSlice(array, n, n2));
    }
    
    public Histogram(int[][] normalizeMatrix) {
        this.bitMetric = new Histogram.BitMetric() {
            @Override
            public double getBitLength(int n) {
                return Histogram.this.getBitLength(n);
            }
        };
        normalizeMatrix = this.normalizeMatrix(normalizeMatrix);
        this.matrix = normalizeMatrix;
        int n = 0;
        int totalWeight = 0;
        for (int i = 0; i < normalizeMatrix.length; ++i) {
            int n2 = normalizeMatrix[i].length - 1;
            n += n2;
            totalWeight += normalizeMatrix[i][0] * n2;
        }
        this.totalWeight = totalWeight;
        long[] a = new long[n];
        int n3 = 0;
        for (int j = 0; j < normalizeMatrix.length; ++j) {
            for (int k = 1; k < normalizeMatrix[j].length; ++k) {
                a[n3++] = ((long)normalizeMatrix[j][k] << 32 | (0xFFFFFFFFL & (long)normalizeMatrix[j][0]));
            }
        }
        assert n3 == a.length;
        Arrays.sort(a);
        int[][] table = makeTable(a);
        this.values = table[1];
        this.counts = table[0];
        assert !(!this.assertWellFormed((int[])null));
        return;
    }
    
    public int[][] getMatrix() {
        return this.matrix;
    }
    
    public int getRowCount() {
        return this.matrix.length;
    }
    
    public int getRowFrequency(int n) {
        return this.matrix[n][0];
    }
    
    public int getRowLength(int n) {
        return this.matrix[n].length - 1;
    }
    
    public int getRowValue(int n, int n2) {
        return this.matrix[n][n2 + 1];
    }
    
    public int getRowWeight(int n) {
        return this.getRowFrequency(n) * this.getRowLength(n);
    }
    
    public int getTotalWeight() {
        return this.totalWeight;
    }
    
    public int getTotalLength() {
        return this.values.length;
    }
    
    public int[] getAllValues() {
        return this.values;
    }
    
    public int[] getAllFrequencies() {
        return this.counts;
    }
    
    public int getFrequency(int key) {
        int binarySearch = Arrays.binarySearch(this.values, key);
        if (binarySearch < 0) {
            return 0;
        }
        assert this.values[binarySearch] == key;
        return this.counts[binarySearch];
    }
    
    public double getBitLength(int n) {
        return -Math.log((double)this.getFrequency(n) / (double)this.getTotalWeight()) / Histogram.log2;
    }
    
    public double getRowBitLength(int n) {
        return -Math.log((double)this.getRowFrequency(n) / (double)this.getTotalWeight()) / Histogram.log2;
    }
    
    public Histogram.BitMetric getBitMetric() {
        return this.bitMetric;
    }
    
    public double getBitLength() {
        double n = 0.0;
        for (int i = 0; i < this.matrix.length; ++i) {
            n += this.getRowBitLength(i) * (double)this.getRowWeight(i);
        }
        assert 0.1 > Math.abs(n - this.getBitLength(this.bitMetric));
        return n;
    }
    
    public double getBitLength(Histogram.BitMetric bitMetric) {
        double n = 0.0;
        for (int i = 0; i < this.matrix.length; ++i) {
            for (int j = 1; j < this.matrix[i].length; ++j) {
                n += (double)this.matrix[i][0] * bitMetric.getBitLength(this.matrix[i][j]);
            }
        }
        return n;
    }
    
    private static double round(double n, double n2) {
        return (double)Math.round(n * n2) / n2;
    }
    
    public int[][] normalizeMatrix(int[][] array) {
        long[] a = new long[array.length];
        for (int i = 0; i < array.length; ++i) {
            if (array[i].length > 1) {
                int n = array[i][0];
                if (n > 0) {
                    a[i] = ((long)n << 32 | (long)i);
                }
            }
        }
        Arrays.sort(a);
        int[][] array2 = new int[array.length][];
        int n2 = -1;
        int j = 0;
        int n3 = 0;
        int n4 = 0;
        while (true) {
            Label_0451: {
                int[] array3;
                if (n4 < array.length) {
                    long n5 = a[a.length - n4 - 1];
                    if (n5 == 0L) {
                        break Label_0451;
                    }
                    array3 = array[(int)n5];
                    assert n5 >>> 32 == (long)array3[0];
                }
                else {
                    array3 = new int[] { -1 };
                }
                if (array3[0] != n2 && n3 > j) {
                    int n6 = 0;
                    while (j < n3) {
                        int[] array4 = array2[j];
                        assert array4[0] == n2;
                        n6 += array4.length - 1;
                        ++j;
                    }
                    int[] a2 = new int[1 + n6];
                    a2[0] = n2;
                    int n7 = 1;
                    while (j < n3) {
                        int[] array5 = array2[j];
                        assert array5[0] == n2;
                        System.arraycopy(array5, 1, a2, n7, array5.length - 1);
                        n7 += array5.length - 1;
                        ++j;
                    }
                    if (!isSorted(a2, 1, true)) {
                        Arrays.sort(a2, 1, a2.length);
                        int n8 = 2;
                        for (int k = 2; k < a2.length; ++k) {
                            if (a2[k] != a2[k - 1]) {
                                a2[n8++] = a2[k];
                            }
                        }
                        if (n8 < a2.length) {
                            int[] array6 = new int[n8];
                            System.arraycopy(a2, 0, array6, 0, n8);
                            a2 = array6;
                        }
                    }
                    array2[j++] = a2;
                    n3 = j;
                }
                if (n4 == array.length) {
                    assert j == n3;
                    array = array2;
                    if (j < array.length) {
                        int[][] array7 = new int[j][];
                        System.arraycopy(array, 0, array7, 0, j);
                        array = array7;
                        return array;
                    }
                    return array;
                }
                else {
                    n2 = array3[0];
                    array2[n3++] = array3;
                }
            }
            ++n4;
        }
    }
    
    public String[] getRowTitles(String str) {
        int totalLength = this.getTotalLength();
        int totalWeight = this.getTotalWeight();
        String[] array = new String[this.matrix.length];
        int i = 0;
        int j = 0;
        int k = 0;
        while (k < this.matrix.length) {
            int rowFrequency = this.getRowFrequency(k);
            int rowLength = this.getRowLength(k);
            i += this.getRowWeight(k);
            j += rowLength;
            long lng = ((long)i * 100L + (long)(totalWeight / 2)) / (long)totalWeight;
            long lng2 = ((long)j * 100L + (long)(totalLength / 2)) / (long)totalLength;
            double rowBitLength = this.getRowBitLength(k);
            assert 0.1 > Math.abs(rowBitLength - this.getBitLength(this.matrix[k][1]));
            array[k] = str + "[" + k + "] len=" + round(rowBitLength, 10.0) + " (" + rowFrequency + "*[" + rowLength + "]) (" + i + ":" + lng + "%) [" + j + ":" + lng2 + "%]";
            ++k;
        }
        return array;
    }
    
    public void print(PrintStream printStream) {
        this.print("hist", printStream);
    }
    
    public void print(String s, PrintStream printStream) {
        this.print(s, this.getRowTitles(s), printStream);
    }
    
    public void print(String str, String[] array, PrintStream printStream) {
        int totalLength = this.getTotalLength();
        int totalWeight = this.getTotalWeight();
        double bitLength = this.getBitLength();
        String string = str + " len=" + round(bitLength, 10.0) + " avgLen=" + round(bitLength / (double)totalWeight, 10.0) + " weight(" + totalWeight + ") unique[" + totalLength + "] avgWeight(" + round((double)totalWeight / (double)totalLength, 100.0) + ")";
        if (array == null) {
            printStream.println(string);
        }
        else {
            printStream.println(string + " {");
            StringBuffer x = new StringBuffer();
            for (int i = 0; i < this.matrix.length; ++i) {
                x.setLength(0);
                x.append("  ").append(array[i]).append(" {");
                for (int j = 1; j < this.matrix[i].length; ++j) {
                    x.append(" ").append(this.matrix[i][j]);
                }
                x.append(" }");
                printStream.println(x);
            }
            printStream.println("}");
        }
    }
    
    private static int[][] makeMatrix(long[] a) {
        Arrays.sort(a);
        int[] array = new int[a.length];
        for (int i = 0; i < array.length; ++i) {
            array[i] = (int)(a[i] >>> 32);
        }
        long[] computeHistogram2Col = computeHistogram2Col(array);
        int[][] array2 = new int[computeHistogram2Col.length][];
        int n = 0;
        int n2 = 0;
        int[] array3;
        for (int length = array2.length; --length >= 0; array2[length] = array3) {
            long n3 = computeHistogram2Col[n2++];
            int n4 = (int)n3;
            int n5 = (int)(n3 >>> 32);
            array3 = new int[1 + n5];
            array3[0] = n4;
            int j = 0;
            while (j < n5) {
                long n6 = a[n++];
                assert n6 >>> 32 == (long)n4;
                array3[1 + j] = (int)n6;
                ++j;
            }
        }
        assert n == a.length;
        return array2;
    }
    
    private static int[][] makeTable(long[] array) {
        int[][] array2 = new int[2][array.length];
        for (int i = 0; i < array.length; ++i) {
            array2[0][i] = (int)array[i];
            array2[1][i] = (int)(array[i] >>> 32);
        }
        return array2;
    }
    
    private static long[] computeHistogram2Col(int[] array) {
        switch (array.length) {
            case 0: {
                return new long[0];
            }
            case 1: {
                return new long[] { 0x100000000L | (0xFFFFFFFFL & (long)array[0]) };
            }
            default: {
                long[] array2 = null;
                int n = 1;
                while (true) {
                    int n2 = -1;
                    int n3 = ~array[0];
                    int n4 = 0;
                    for (int i = 0; i <= array.length; ++i) {
                        int n5;
                        if (i < array.length) {
                            n5 = array[i];
                        }
                        else {
                            n5 = ~n3;
                        }
                        if (n5 == n3) {
                            ++n4;
                        }
                        else {
                            if (n == 0 && n4 != 0) {
                                array2[n2] = ((long)n4 << 32 | (0xFFFFFFFFL & (long)n3));
                            }
                            n3 = n5;
                            n4 = 1;
                            ++n2;
                        }
                    }
                    if (n != 0) {
                        array2 = new long[n2];
                        n = 0;
                    }
                    else {
                        break;
                    }
                }
                return array2;
            }
        }
    }
    
    private static int[][] regroupHistogram(int[][] array, int[] array2) {
        long n = 0L;
        for (int i = 0; i < array.length; ++i) {
            n += (long)(array[i].length - 1);
        }
        long n2 = 0L;
        for (int j = 0; j < array2.length; ++j) {
            n2 += (long)array2[j];
        }
        if (n2 > n) {
            int length = array2.length;
            int k = 0;
            while (k < array2.length) {
                if (n < (long)array2[k]) {
                    int[] array4 = new int[k + 1];
                    System.arraycopy(array2, 0, array4, 0, k + 1);
                    array2 = array4;
                    array2[k] = (int)n;
                    break;
                }
                n -= (long)array2[k];
                ++k;
            }
        }
        else {
            long n3 = n - n2;
            int[] array3 = new int[array2.length + 1];
            System.arraycopy(array2, 0, array3, 0, array2.length);
            array3[array2.length] = (int)n3;
            array2 = array3;
        }
        int[][] array5 = new int[array2.length][];
        int n4 = 0;
        int l = 1;
        int n5 = array[n4].length;
        int n7;
        int[] a;
        long n8;
        for (int n6 = 0; n6 < array2.length; a[0] = (int)((n8 + (long)(n7 / 2)) / (long)n7), ++n6) {
            n7 = array2[n6];
            a = new int[1 + n7];
            n8 = 0L;
            array5[n6] = a;
            int n10;
            for (int n9 = 1; n9 < a.length; n5 -= n10, n9 += n10) {
                n10 = a.length - n9;
                while (l == n5) {
                    l = 1;
                    n5 = array[++n4].length;
                }
                if (n10 > n5 - l) {
                    n10 = n5 - l;
                }
                n8 += (long)array[n4][0] * (long)n10;
                System.arraycopy(array[n4], n5 - n10, a, n9, n10);
            }
            Arrays.sort(a, 1, a.length);
        }
        assert l == n5;
        assert n4 == array.length - 1;
        return array5;
    }
    
    public static Histogram makeByteHistogram(InputStream inputStream) throws IOException {
        byte[] b = new byte[4096];
        int[] array = new int[256];
        int read;
        while ((read = inputStream.read(b)) > 0) {
            for (int i = 0; i < read; ++i) {
                int n = b[i] & 0xFF;
                ++array[n];
            }
        }
        int[][] array2 = new int[256][2];
        for (int j = 0; j < array.length; ++j) {
            array2[j][0] = array[j];
            array2[j][1] = j;
        }
        return new Histogram(array2);
    }
    
    private static int[] sortedSlice(int[] array, int n, int n2) {
        if (n == 0 && n2 == array.length && isSorted(array, 0, false)) {
            return array;
        }
        int[] a = new int[n2 - n];
        System.arraycopy(array, n, a, 0, a.length);
        Arrays.sort(a);
        return a;
    }
    
    private static boolean isSorted(int[] array, int n, boolean b) {
        int i = n + 1;
        while (i < array.length) {
            if (!(b ? (array[i - 1] < array[i]) : (array[i - 1] <= array[i]))) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    private static int[] maybeSort(int[] a) {
        if (!isSorted(a, 0, false)) {
            a = (int[])a.clone();
            Arrays.sort(a);
            return a;
        }
        return a;
    }
    
    private boolean assertWellFormed(int[] array) {
        return true;
    }
    
    static {
        Histogram.log2 = Math.log(2.0);
    }
    
    public interface BitMetric
    {
        public abstract double getBitLength(int p0);
    }
}
